@@ -1,5 +1,7 @@
 import pyvista as pv
 import numpy as np
+import re
+
 from LoopStructural.datatypes import VectorPoints, ValuePoints
 from LoopStructural.modelling.features import BaseFeature, StructuralFrame
 
@@ -7,7 +9,7 @@ from LoopStructural.modelling.features.fault import FaultSegment
 from LoopStructural.datatypes import BoundingBox
 from LoopStructural import GeologicalModel
 from LoopStructural.utils import getLogger
-from typing import Union, Optional, List
+from typing import Callable, Union, Optional, List
 from ._colours import random_colour
 
 logger = getLogger(__name__)
@@ -25,20 +27,38 @@ class Loop3DView(pv.Plotter):
         background : str, optional
             colour for the background, by default 'white'
         """
+        if 'shape' in kwargs:
+            logger.warning('shape argument is not used in Loop3DView')
+            kwargs.pop('shape')
         super().__init__(*args, **kwargs)
         self.set_background(background)
         self.model = model
         self.objects = {}
 
+    def subplot(self, *args, **kwargs):
+        logger.warning('subplot is not supported in Loop3DView')
+        return self
+
     def add_mesh(self, *args, **kwargs):
-        if 'name' not in kwargs:
+        if 'name' not in kwargs or kwargs['name'] is None:
             name = 'unnamed_object'
-            name = self.increment_name(name)
-            name.replace(' ', '_')  # spaces seem to cause issues for vue
             kwargs['name'] = name
             logger.warning(
                 f'No name provided, using {name}. Pass name argument to add_mesh to remove this error'
             )
+        kwargs['name'] = kwargs['name'].replace(' ', '_')
+        kwargs['name'] = re.sub(r'[^a-zA-Z0-9_$]', '_', kwargs['name'])
+        if kwargs['name'][0].isdigit():
+            kwargs['name'] = 'ls_' + kwargs['name']
+        if kwargs['name'][0] == '_':
+            kwargs['name'] = 'ls' + kwargs['name']
+        kwargs['name'] = self.increment_name(kwargs['name'])
+        if '__opacity' in kwargs['name']:
+            raise ValueError('Cannot use __opacity in name')
+        if '__visibility' in kwargs['name']:
+            raise ValueError('Cannot use __visibility in name')
+        if '__control_visibility' in kwargs['name']:
+            raise ValueError('Cannot use __control_visibility in name')
         return super().add_mesh(*args, **kwargs)
 
     def increment_name(self, name):
@@ -62,6 +82,22 @@ class Loop3DView(pv.Plotter):
             raise ValueError("No model provided")
         return model
 
+    def _get_vector_scale(self, scale: Optional[Union[float, int]]) -> float:
+        autoscale = 0.0
+        if self.model is not None:
+            # automatically scale vector data to be 5% of the bounding box length
+            autoscale = self.model.bounding_box.length.max() * 0.05
+        if scale is None:
+            scale = autoscale
+        else:
+            scale = scale * autoscale
+        if scale > 10 * autoscale:
+            logger.warning(
+                "Vector scale magnitude greater than half of the model bounding box length, is this correct?"
+            )
+
+        return scale
+
     def plot_surface(
         self,
         geological_feature: BaseFeature,
@@ -76,6 +112,7 @@ class Loop3DView(pv.Plotter):
         scalar_bar: bool = False,
         slicer: bool = False,
         name: Optional[str] = None,
+        bounding_box: Optional[BoundingBox] = None,
     ):
         """Add an isosurface of a geological feature to the model
 
@@ -111,7 +148,7 @@ class Loop3DView(pv.Plotter):
             name = geological_feature.name + '_surfaces'
         name = self.increment_name(name)  # , 'surface')
 
-        surfaces = geological_feature.surfaces(value)
+        surfaces = geological_feature.surfaces(value, bounding_box=bounding_box)
         meshes = []
         for surface in surfaces:
             s = surface.vtk()
@@ -338,18 +375,17 @@ class Loop3DView(pv.Plotter):
                 if g == "faults":
                     continue
                 for v in model.stratigraphic_column[g].values():
-                    if not isinstance(v['colour'], str):
+                    colour = v.get("colour", None)
+                    if not isinstance(colour, str):
                         try:
-                            v['colour'] = colors.to_hex(v['colour'])
+                            v['colour'] = colors.to_hex(colour)
                         except ValueError:
-                            logger.warning(
-                                f"Cannot convert colour {v['colour']} to hex, using default"
-                            )
+                            logger.warning(f"Cannot convert colour {colour} to hex, using default")
                             v['colour'] = random_colour()
                     data.append((v["id"], v["colour"]))
                     colours.append(v["colour"])
                     boundaries.append(v["id"])  # print(u,v)
-            cmap = colors.ListedColormap(colours).colors
+            cmap = colors.ListedColormap(colours)
         except ImportError:
             logger.warning("Cannot use predefined colours as I can't import matplotlib")
             cmap = "tab20"
@@ -399,7 +435,7 @@ class Loop3DView(pv.Plotter):
             strati_surfaces = []
             surfaces = model.get_stratigraphic_surfaces()
             if cmap is None:
-                cmap = self._build_stratigraphic_cmap(model)
+                cmap = self._build_stratigraphic_cmap(model).colors
             for s in surfaces:
                 strati_surfaces.append(s.vtk())
             if name is None:
@@ -433,8 +469,12 @@ class Loop3DView(pv.Plotter):
     def plot_vector_field(
         self,
         geological_feature: BaseFeature,
-        scale: float = 1.0,
+        scale: Optional[float] = None,
         name: Optional[str] = None,
+        geom='arrow',
+        scalars: Optional[np.ndarray] = None,
+        normalise: bool = False,
+        scale_function: Optional[Callable[[np.ndarray], np.ndarray]] = None,
         pyvista_kwargs: dict = {},
     ) -> pv.Actor:
         """Plot a vector field
@@ -459,16 +499,29 @@ class Loop3DView(pv.Plotter):
             name = geological_feature.name + '_vector_field'
         name = self.increment_name(name)  # , 'vector_field')
         vectorfield = geological_feature.vector_field()
-        return self.add_mesh(vectorfield.vtk(scale=scale), name=name, **pyvista_kwargs)
+        scale = self._get_vector_scale(scale)
+        return self.add_mesh(
+            vectorfield.vtk(
+                scale=scale,
+                geom=geom,
+                normalise=normalise,
+                scalars=scalars,
+                scale_function=scale_function,
+            ),
+            name=name,
+            **pyvista_kwargs,
+        )
 
     def plot_data(
         self,
         feature: Union[BaseFeature, StructuralFrame],
         value: bool = True,
         vector: bool = True,
-        scale: Union[float, int] = 10,
+        scale: Optional[Union[float, int]] = None,
         geom: str = "arrow",
         name: Optional[str] = None,
+        scalars: Optional[np.ndarray] = None,
+        normalise: bool = True,
         pyvista_kwargs: dict = {},
     ) -> List[pv.Actor]:
         """Add the data associated with a feature to the plotter
@@ -482,11 +535,13 @@ class Loop3DView(pv.Plotter):
         vector : bool, optional
             whether to plot vector data, by default True
         scale : Union[float, int], optional
-            vector scale, by default 10
+            vector scale, by default 1
         geom : str, optional
             vector glyph, by default "arrow"
         name : Optional[str], optional
             name to use in object list, by default None
+        normalise: bool, optional
+            normalise the vectors to be unit norm, by default True
         pyvista_kwargs : dict, optional
             additional kwargs to pass to pyvista add_mesh, by default {}
 
@@ -494,11 +549,19 @@ class Loop3DView(pv.Plotter):
         -------
         List[pv.Actor]
             list of actors added to the pv plotter
+
+        Notes
+        ------
+        When plotting a vector the bounding box is used to scale the vectors. By default
+        the length of the arrows will be 5% of the bounding box. The scale parameter is a
+        multiplier for this value. If you sent normalise to False the vectors will not be normalised
+
         """
         if issubclass(type(feature), BaseFeature):
             feature = [feature]
-
+        scale = self._get_vector_scale(scale)
         actors = []
+        bb = self.model.bounding_box if self.model is not None else None
         for f in feature:
             for d in f.get_data():
                 if isinstance(d, ValuePoints):
@@ -508,7 +571,11 @@ class Loop3DView(pv.Plotter):
                         else:
                             object_name = f'{d.name}_values_{name}'
                         object_name = self.increment_name(object_name)  # , 'values')
-                        actors.append(self.add_mesh(d.vtk(), name=object_name, **pyvista_kwargs))
+                        actors.append(
+                            self.add_mesh(
+                                d.vtk(scalars=scalars), name=object_name, **pyvista_kwargs
+                            )
+                        )
                 if isinstance(d, VectorPoints):
                     if vector:
                         if name is None:
@@ -518,7 +585,16 @@ class Loop3DView(pv.Plotter):
                         object_name = self.increment_name(object_name)  # , 'vectors')
                         actors.append(
                             self.add_mesh(
-                                d.vtk(geom=geom, scale=scale), name=name, **pyvista_kwargs
+                                d.vtk(
+                                    geom=geom,
+                                    scale=scale,
+                                    scalars=scalars,
+                                    bb=bb,
+                                    tolerance=None,
+                                    normalise=normalise,
+                                ),
+                                name=name,
+                                **pyvista_kwargs,
                             )
                         )
         return actors
@@ -535,8 +611,9 @@ class Loop3DView(pv.Plotter):
         slip_vector: bool = True,
         displacement_scale_vector: bool = True,
         fault_volume: bool = True,
-        vector_scale: Union[float, int] = 200,
+        vector_scale: Optional[Union[float, int]] = None,
         name: Optional[str] = None,
+        geom: str = "arrow",
         pyvista_kwargs: dict = {},
     ) -> List[pv.Actor]:
         """Plot a fault including the surface, slip vector and displacement volume
@@ -581,17 +658,11 @@ class Loop3DView(pv.Plotter):
                 vector_name = f'{fault.name}_vector_{name}'
             vector_name = self.increment_name(vector_name)
 
-            vectorfield = fault[1].vector_field()
+            vectorfield = fault.vector_field()
+            vector_scale = self._get_vector_scale(vector_scale)
             actors.append(
                 self.add_mesh(
-                    vectorfield.vtk(
-                        scale=vector_scale,
-                        scale_function=(
-                            fault.displacementfeature.evaluate_value
-                            if displacement_scale_vector
-                            else None
-                        ),
-                    ),
+                    vectorfield.vtk(scale=vector_scale, normalise=False),
                     name=vector_name,
                     **pyvista_kwargs,
                 )
@@ -602,12 +673,44 @@ class Loop3DView(pv.Plotter):
             else:
                 volume_name = f'{fault.name}_volume_{name}'
             volume = fault.displacementfeature.scalar_field()
-            volume = volume.vtk().threshold(0.0)
+            volume = volume.vtk().threshold([-1.0, 1.0])
+            if geom == "arrow":
+                geom = pv.Arrow()
+            elif geom == "disc":
+                geom = pv.Disc()
+                geom = geom.rotate_y(90)
+            else:
+                raise ValueError(f"Unknown glyph type {geom}")
             actors.append(self.add_mesh(volume, name=volume_name, **pyvista_kwargs))
         if len(actors) == 0:
             logger.warning(f"Nothing added to plot for {fault.name}")
         return actors
 
+    def plot_fault_ellipsoid(
+        self, fault: FaultSegment, name: Optional[str] = None, pyvista_kwargs: dict = {}
+    ) -> pv.Actor:
+        """Plot the fault ellipsoid
+
+        Parameters
+        ----------
+        fault : FaultSegment
+            the fault to plot
+        name : Optional[str], optional
+            name of the object for pyvista, by default None
+        pyvista_kwargs : dict, optional
+            additional kwargs for the pyvista plotter, by default {}
+
+        Returns
+        -------
+        pv.Actor
+            actor added to the plot
+        """
+        if name is None:
+            name = fault.name + '_ellipsoid'
+        name = self.increment_name(name)
+        ellipsoid = fault.fault_ellipsoid()
+        return self.add_mesh(ellipsoid.vtk(), name=name, **pyvista_kwargs)
+
     def rotate(self, angles: np.ndarray):
         """Rotate the camera by the given angles
         order is roll, azimuth, elevation as defined by
